# -*- coding: utf-8 -*-
# ****************************************************************
# aitk.algorithms: Python algorithms for AI
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.algorithms
#
# ****************************************************************

from .ga import GeneticAlgorithm
from .novelty import NoveltySearch
